@@ -1,11 +1,12 @@
 // Test GoogleCloudStorage filesystem interface
+
 package googlecloudstorage_test
 
 import (
 	"testing"
 
-	"github.com/ncw/rclone/backend/googlecloudstorage"
-	"github.com/ncw/rclone/fstest/fstests"
+	"github.com/rclone/rclone/backend/googlecloudstorage"
+	"github.com/rclone/rclone/fstest/fstests"
 )
 
 // TestIntegration runs integration tests against the remote
