@@ -1,4 +1,5 @@
-//+build windows
+//go:build windows
+// +build windows
 
 package local
 
@@ -7,7 +8,7 @@ import (
 	"syscall"
 	"time"
 
-	"github.com/ncw/rclone/fs"
+	"github.com/rclone/rclone/fs"
 )
 
 const (
