@@ -1,14 +1,16 @@
-// +build !plan9
+//go:build !plan9 && !js
+// +build !plan9,!js
 
+// Package cachestats provides the cachestats command.
 package cachestats
 
 import (
 	"encoding/json"
 	"fmt"
 
-	"github.com/pkg/errors"
 	"github.com/rclone/rclone/backend/cache"
 	"github.com/rclone/rclone/cmd"
+	"github.com/rclone/rclone/fs"
 	"github.com/spf13/cobra"
 )
 
@@ -22,8 +24,10 @@ var commandDefinition = &cobra.Command{
 	Long: `
 Print cache stats for a remote in JSON format
 `,
+	Hidden: true,
 	Run: func(command *cobra.Command, args []string) {
 		cmd.CheckArgs(1, 1, command, args)
+		fs.Logf(nil, `"rclone cachestats" is deprecated, use "rclone backend stats %s" instead`, args[0])
 
 		fsrc := cmd.NewFsSrc(args)
 		cmd.Run(false, false, command, func() error {
@@ -35,7 +39,7 @@ Print cache stats for a remote in JSON format
 					fsCache, ok = unwrap().(*cache.Fs)
 				}
 				if !ok {
-					return errors.Errorf("%s: is not a cache remote", fsrc.Name())
+					return fmt.Errorf("%s: is not a cache remote", fsrc.Name())
 				}
 			}
 			m, err := fsCache.Stats()
