@@ -1,3 +1,4 @@
+// Package cleanup provides the cleanup command.
 package cleanup
 
 import (
@@ -14,7 +15,7 @@ func init() {
 
 var commandDefinition = &cobra.Command{
 	Use:   "cleanup remote:path",
-	Short: `Clean up the remote if possible`,
+	Short: `Clean up the remote if possible.`,
 	Long: `
 Clean up the remote if possible.  Empty the trash or delete old file
 versions. Not supported by all remotes.
