@@ -1,36 +1,32 @@
+// Package cryptcheck provides the cryptcheck command.
 package cryptcheck
 
 import (
 	"context"
+	"fmt"
 
-	"github.com/pkg/errors"
 	"github.com/rclone/rclone/backend/crypt"
 	"github.com/rclone/rclone/cmd"
+	"github.com/rclone/rclone/cmd/check"
 	"github.com/rclone/rclone/fs"
-	"github.com/rclone/rclone/fs/config/flags"
 	"github.com/rclone/rclone/fs/hash"
 	"github.com/rclone/rclone/fs/operations"
 	"github.com/spf13/cobra"
 )
 
-// Globals
-var (
-	oneway = false
-)
-
 func init() {
 	cmd.Root.AddCommand(commandDefinition)
 	cmdFlag := commandDefinition.Flags()
-	flags.BoolVarP(cmdFlag, &oneway, "one-way", "", oneway, "Check one way only, source files must exist on destination")
+	check.AddFlags(cmdFlag)
 }
 
 var commandDefinition = &cobra.Command{
 	Use:   "cryptcheck remote:path cryptedremote:path",
 	Short: `Cryptcheck checks the integrity of a crypted remote.`,
 	Long: `
-rclone cryptcheck checks a remote against a crypted remote.  This is
-the equivalent of running rclone check, but able to check the
-checksums of the crypted remote.
+rclone cryptcheck checks a remote against a [crypted](/crypt/) remote.
+This is the equivalent of running rclone [check](/commands/rclone_check/),
+but able to check the checksums of the crypted remote.
 
 For it to work the underlying remote of the cryptedremote must support
 some kind of checksum.
@@ -50,11 +46,7 @@ the files in remote:path.
     rclone cryptcheck remote:path encryptedremote:path
 
 After it has run it will log the status of the encryptedremote:.
-
-If you supply the --one-way flag, it will only check that files in source
-match the files in destination, not the other way around. Meaning extra files in
-destination that are not in the source will not trigger an error.
-`,
+` + check.FlagsHelp,
 	Run: func(command *cobra.Command, args []string) {
 		cmd.CheckArgs(2, 2, command, args)
 		fsrc, fdst := cmd.NewFsSrcDst(args)
@@ -69,50 +61,50 @@ func cryptCheck(ctx context.Context, fdst, fsrc fs.Fs) error {
 	// Check to see fcrypt is a crypt
 	fcrypt, ok := fdst.(*crypt.Fs)
 	if !ok {
-		return errors.Errorf("%s:%s is not a crypt remote", fdst.Name(), fdst.Root())
+		return fmt.Errorf("%s:%s is not a crypt remote", fdst.Name(), fdst.Root())
 	}
 	// Find a hash to use
 	funderlying := fcrypt.UnWrap()
 	hashType := funderlying.Hashes().GetOne()
 	if hashType == hash.None {
-		return errors.Errorf("%s:%s does not support any hashes", funderlying.Name(), funderlying.Root())
+		return fmt.Errorf("%s:%s does not support any hashes", funderlying.Name(), funderlying.Root())
 	}
 	fs.Infof(nil, "Using %v for hash comparisons", hashType)
 
+	opt, close, err := check.GetCheckOpt(fsrc, fcrypt)
+	if err != nil {
+		return err
+	}
+	defer close()
+
 	// checkIdentical checks to see if dst and src are identical
 	//
 	// it returns true if differences were found
 	// it also returns whether it couldn't be hashed
-	checkIdentical := func(ctx context.Context, dst, src fs.Object) (differ bool, noHash bool) {
+	opt.Check = func(ctx context.Context, dst, src fs.Object) (differ bool, noHash bool, err error) {
 		cryptDst := dst.(*crypt.Object)
 		underlyingDst := cryptDst.UnWrap()
 		underlyingHash, err := underlyingDst.Hash(ctx, hashType)
 		if err != nil {
-			fs.CountError(err)
-			fs.Errorf(dst, "Error reading hash from underlying %v: %v", underlyingDst, err)
-			return true, false
+			return true, false, fmt.Errorf("error reading hash from underlying %v: %w", underlyingDst, err)
 		}
 		if underlyingHash == "" {
-			return false, true
+			return false, true, nil
 		}
 		cryptHash, err := fcrypt.ComputeHash(ctx, cryptDst, src, hashType)
 		if err != nil {
-			fs.CountError(err)
-			fs.Errorf(dst, "Error computing hash: %v", err)
-			return true, false
+			return true, false, fmt.Errorf("error computing hash: %w", err)
 		}
 		if cryptHash == "" {
-			return false, true
+			return false, true, nil
 		}
 		if cryptHash != underlyingHash {
-			err = errors.Errorf("hashes differ (%s:%s) %q vs (%s:%s) %q", fdst.Name(), fdst.Root(), cryptHash, fsrc.Name(), fsrc.Root(), underlyingHash)
-			fs.CountError(err)
+			err = fmt.Errorf("hashes differ (%s:%s) %q vs (%s:%s) %q", fdst.Name(), fdst.Root(), cryptHash, fsrc.Name(), fsrc.Root(), underlyingHash)
 			fs.Errorf(src, err.Error())
-			return true, false
+			return true, false, nil
 		}
-		fs.Debugf(src, "OK")
-		return false, false
+		return false, false, nil
 	}
 
-	return operations.CheckFn(ctx, fcrypt, fsrc, checkIdentical, oneway)
+	return operations.CheckFn(ctx, opt)
 }
