@@ -1,19 +1,24 @@
 package cmd
 
 import (
+	"context"
 	"fmt"
 	"log"
 	"os"
 	"regexp"
+	"sort"
 	"strings"
 
 	"github.com/rclone/rclone/fs"
 	"github.com/rclone/rclone/fs/config/configflags"
 	"github.com/rclone/rclone/fs/filter/filterflags"
+	"github.com/rclone/rclone/fs/log/logflags"
 	"github.com/rclone/rclone/fs/rc/rcflags"
 	"github.com/rclone/rclone/lib/atexit"
 	"github.com/spf13/cobra"
 	"github.com/spf13/pflag"
+	"golang.org/x/text/cases"
+	"golang.org/x/text/language"
 )
 
 // Root is the main rclone command
@@ -46,10 +51,11 @@ __rclone_custom_func() {
         else
             __rclone_init_completion -n : || return
         fi
+	local rclone=(command rclone --ask-password=false)
         if [[ $cur != *:* ]]; then
             local ifs=$IFS
             IFS=$'\n'
-            local remotes=($(command rclone listremotes))
+            local remotes=($("${rclone[@]}" listremotes 2> /dev/null))
             IFS=$ifs
             local remote
             for remote in "${remotes[@]}"; do
@@ -68,7 +74,7 @@ __rclone_custom_func() {
             fi
             local ifs=$IFS
             IFS=$'\n'
-            local lines=($(rclone lsf "${cur%%:*}:$prefix" 2>/dev/null))
+            local lines=($("${rclone[@]}" lsf "${cur%%:*}:$prefix" 2> /dev/null))
             IFS=$ifs
             local line
             for line in "${lines[@]}"; do
@@ -162,12 +168,14 @@ func runRoot(cmd *cobra.Command, args []string) {
 // setupRootCommand sets default usage, help, and error handling for
 // the root command.
 //
-// Helpful example: http://rtfcode.com/xref/moby-17.03.2-ce/cli/cobra.go
+// Helpful example: https://github.com/moby/moby/blob/master/cli/cobra.go
 func setupRootCommand(rootCmd *cobra.Command) {
+	ci := fs.GetConfig(context.Background())
 	// Add global flags
-	configflags.AddFlags(pflag.CommandLine)
+	configflags.AddFlags(ci, pflag.CommandLine)
 	filterflags.AddFlags(pflag.CommandLine)
 	rcflags.AddFlags(pflag.CommandLine)
+	logflags.AddFlags(pflag.CommandLine)
 
 	Root.Run = runRoot
 	Root.Flags().BoolVarP(&version, "version", "V", false, "Print the version number")
@@ -242,7 +250,6 @@ Use "rclone help backends" for a list of supported services.
 var docFlagsTemplate = `---
 title: "Global Flags"
 description: "Rclone Global Flags"
-date: "YYYY-MM-DD"
 ---
 
 # Global Flags
@@ -295,7 +302,7 @@ func showBackend(name string) {
 	var standardOptions, advancedOptions fs.Options
 	done := map[string]struct{}{}
 	for _, opt := range backend.Options {
-		// Skip if done already (eg with Provider options)
+		// Skip if done already (e.g. with Provider options)
 		if _, doneAlready := done[opt.Name]; doneAlready {
 			continue
 		}
@@ -311,7 +318,8 @@ func showBackend(name string) {
 			optionsType = "advanced"
 			continue
 		}
-		fmt.Printf("### %s Options\n\n", strings.Title(optionsType))
+		optionsType = cases.Title(language.Und, cases.NoLower).String(optionsType)
+		fmt.Printf("### %s options\n\n", optionsType)
 		fmt.Printf("Here are the %s options specific to %s (%s).\n\n", optionsType, backend.Name, backend.Description)
 		optionsType = "advanced"
 		for _, opt := range opts {
@@ -322,12 +330,32 @@ func showBackend(name string) {
 			}
 			fmt.Printf("#### --%s%s\n\n", opt.FlagName(backend.Prefix), shortOpt)
 			fmt.Printf("%s\n\n", opt.Help)
+			if opt.IsPassword {
+				fmt.Printf("**NB** Input to this must be obscured - see [rclone obscure](/commands/rclone_obscure/).\n\n")
+			}
+			fmt.Printf("Properties:\n\n")
 			fmt.Printf("- Config:      %s\n", opt.Name)
 			fmt.Printf("- Env Var:     %s\n", opt.EnvVarName(backend.Prefix))
+			if opt.Provider != "" {
+				fmt.Printf("- Provider:    %s\n", opt.Provider)
+			}
 			fmt.Printf("- Type:        %s\n", opt.Type())
-			fmt.Printf("- Default:     %s\n", quoteString(opt.GetValue()))
+			defaultValue := opt.GetValue()
+			// Default value and Required are related: Required means option must
+			// have a value, but if there is a default then a value does not have
+			// to be explicitly set and then Required makes no difference.
+			if defaultValue != "" {
+				fmt.Printf("- Default:     %s\n", quoteString(defaultValue))
+			} else {
+				fmt.Printf("- Required:    %v\n", opt.Required)
+			}
+			// List examples / possible choices
 			if len(opt.Examples) > 0 {
-				fmt.Printf("- Examples:\n")
+				if opt.Exclusive {
+					fmt.Printf("- Choices:\n")
+				} else {
+					fmt.Printf("- Examples:\n")
+				}
 				for _, ex := range opt.Examples {
 					fmt.Printf("    - %s\n", quoteString(ex.Value))
 					for _, line := range strings.Split(ex.Help, "\n") {
@@ -338,4 +366,28 @@ func showBackend(name string) {
 			fmt.Printf("\n")
 		}
 	}
+	if backend.MetadataInfo != nil {
+		fmt.Printf("### Metadata\n\n")
+		fmt.Printf("%s\n\n", strings.TrimSpace(backend.MetadataInfo.Help))
+		if len(backend.MetadataInfo.System) > 0 {
+			fmt.Printf("Here are the possible system metadata items for the %s backend.\n\n", backend.Name)
+			keys := []string{}
+			for k := range backend.MetadataInfo.System {
+				keys = append(keys, k)
+			}
+			sort.Strings(keys)
+			fmt.Printf("| Name | Help | Type | Example | Read Only |\n")
+			fmt.Printf("|------|------|------|---------|-----------|\n")
+			for _, k := range keys {
+				v := backend.MetadataInfo.System[k]
+				ro := "N"
+				if v.ReadOnly {
+					ro = "**Y**"
+				}
+				fmt.Printf("| %s | %s | %s | %s | %s |\n", k, v.Help, v.Type, v.Example, ro)
+			}
+			fmt.Printf("\n")
+		}
+		fmt.Printf("See the [metadata](/docs/#metadata) docs for more info.\n\n")
+	}
 }
