@@ -1,3 +1,4 @@
+//go:build ignore
 // +build ignore
 
 // Read two files with lots of seeking to stress test the seek code
@@ -16,7 +17,7 @@ import (
 
 var (
 	// Flags
-	iterations   = flag.Int("n", 1E6, "Iterations to try")
+	iterations   = flag.Int("n", 1e6, "Iterations to try")
 	maxBlockSize = flag.Int("b", 1024*1024, "Max block size to read")
 )
 
