@@ -1,7 +1,7 @@
+//go:build !plan9 && !js
+// +build !plan9,!js
+
 // Package ncdu implements a text based user interface for exploring a remote
-
-//+build !plan9,!solaris
-
 package ncdu
 
 import (
@@ -13,12 +13,12 @@ import (
 	"strings"
 
 	"github.com/atotto/clipboard"
+	"github.com/gdamore/tcell/v2/termbox"
 	runewidth "github.com/mattn/go-runewidth"
-	termbox "github.com/nsf/termbox-go"
-	"github.com/pkg/errors"
 	"github.com/rclone/rclone/cmd"
 	"github.com/rclone/rclone/cmd/ncdu/scan"
 	"github.com/rclone/rclone/fs"
+	"github.com/rclone/rclone/fs/fspath"
 	"github.com/rclone/rclone/fs/operations"
 	"github.com/spf13/cobra"
 )
@@ -35,23 +35,43 @@ This displays a text based user interface allowing the navigation of a
 remote. It is most useful for answering the question - "What is using
 all my disk space?".
 
-<script src="https://asciinema.org/a/157793.js" id="asciicast-157793" async></script>
+{{< asciinema 157793 >}}
 
 To make the user interface it first scans the entire remote given and
 builds an in memory representation.  rclone ncdu can be used during
 this scanning phase and you will see it building up the directory
 structure as it goes along.
 
-Here are the keys - press '?' to toggle the help on and off
+You can interact with the user interface using key presses,
+press '?' to toggle the help on and off. The supported keys are:
 
     ` + strings.Join(helpText()[1:], "\n    ") + `
 
+Listed files/directories may be prefixed by a one-character flag,
+some of them combined with a description in brackets at end of line.
+These flags have the following meaning:
+
+    e means this is an empty directory, i.e. contains no files (but
+      may contain empty subdirectories)
+    ~ means this is a directory where some of the files (possibly in
+      subdirectories) have unknown size, and therefore the directory
+      size may be underestimated (and average size inaccurate, as it
+      is average of the files with known sizes).
+    . means an error occurred while reading a subdirectory, and
+      therefore the directory size may be underestimated (and average
+      size inaccurate)
+    ! means an error occurred while reading this directory
+
 This an homage to the [ncdu tool](https://dev.yorhel.nl/ncdu) but for
 rclone remotes.  It is missing lots of features at the moment
 but is useful as it stands.
 
-Note that it might take some time to delete big files/folders. The
+Note that it might take some time to delete big files/directories. The
 UI won't respond in the meantime since the deletion is done synchronously.
+
+For a non-interactive listing of the remote, see the
+[tree](/commands/rclone_tree/) command. To just get the total size of
+the remote you can also use the [size](/commands/rclone_size/) command.
 `,
 	Run: func(command *cobra.Command, args []string) {
 		cmd.CheckArgs(1, 1, command, args)
@@ -71,44 +91,54 @@ func helpText() (tr []string) {
 		" ←,h to return",
 		" c toggle counts",
 		" g toggle graph",
-		" n,s,C sort by name,size,count",
+		" a toggle average size in directory",
+		" u toggle human-readable format",
+		" n,s,C,A sort by name,size,count,average size",
 		" d delete file/directory",
+		" v select file/directory",
+		" V enter visual select mode",
+		" D delete selected files/directories",
 	}
 	if !clipboard.Unsupported {
-		tr = append(tr, " y copy current path to clipbard")
+		tr = append(tr, " y copy current path to clipboard")
 	}
 	tr = append(tr, []string{
 		" Y display current path",
-		" ^L refresh screen",
+		" ^L refresh screen (fix screen corruption)",
 		" ? to toggle help on and off",
-		" q/ESC/c-C to quit",
+		" q/ESC/^c to quit",
 	}...)
 	return
 }
 
 // UI contains the state of the user interface
 type UI struct {
-	f              fs.Fs     // fs being displayed
-	fsName         string    // human name of Fs
-	root           *scan.Dir // root directory
-	d              *scan.Dir // current directory being displayed
-	path           string    // path of current directory
-	showBox        bool      // whether to show a box
-	boxText        []string  // text to show in box
-	boxMenu        []string  // box menu options
-	boxMenuButton  int
-	boxMenuHandler func(fs fs.Fs, path string, option int) (string, error)
-	entries        fs.DirEntries // entries of current directory
-	sortPerm       []int         // order to display entries in after sorting
-	invSortPerm    []int         // inverse order
-	dirListHeight  int           // height of listing
-	listing        bool          // whether listing is in progress
-	showGraph      bool          // toggle showing graph
-	showCounts     bool          // toggle showing counts
-	sortByName     int8          // +1 for normal, 0 for off, -1 for reverse
-	sortBySize     int8
-	sortByCount    int8
-	dirPosMap      map[string]dirPos // store for directory positions
+	f                  fs.Fs     // fs being displayed
+	fsName             string    // human name of Fs
+	root               *scan.Dir // root directory
+	d                  *scan.Dir // current directory being displayed
+	path               string    // path of current directory
+	showBox            bool      // whether to show a box
+	boxText            []string  // text to show in box
+	boxMenu            []string  // box menu options
+	boxMenuButton      int
+	boxMenuHandler     func(fs fs.Fs, path string, option int) (string, error)
+	entries            fs.DirEntries // entries of current directory
+	sortPerm           []int         // order to display entries in after sorting
+	invSortPerm        []int         // inverse order
+	dirListHeight      int           // height of listing
+	listing            bool          // whether listing is in progress
+	showGraph          bool          // toggle showing graph
+	showCounts         bool          // toggle showing counts
+	showDirAverageSize bool          // toggle average size
+	humanReadable      bool          // toggle human-readable format
+	visualSelectMode   bool          // toggle visual selection mode
+	sortByName         int8          // +1 for normal, 0 for off, -1 for reverse
+	sortBySize         int8
+	sortByCount        int8
+	sortByAverageSize  int8
+	dirPosMap          map[string]dirPos // store for directory positions
+	selectedEntries    map[string]dirPos // selected entries of current directory
 }
 
 // Where we have got to in the directory listing
@@ -278,24 +308,35 @@ func (u *UI) biggestEntry() (biggest int64) {
 		return
 	}
 	for i := range u.entries {
-		size, _, _, _ := u.d.AttrI(u.sortPerm[i])
-		if size > biggest {
-			biggest = size
+		attrs, _ := u.d.AttrI(u.sortPerm[i])
+		if attrs.Size > biggest {
+			biggest = attrs.Size
 		}
 	}
 	return
 }
 
+// hasEmptyDir returns true if there is empty folder in current listing
+func (u *UI) hasEmptyDir() bool {
+	if u.d == nil {
+		return false
+	}
+	for i := range u.entries {
+		attrs, _ := u.d.AttrI(u.sortPerm[i])
+		if attrs.IsDir && attrs.Count == 0 {
+			return true
+		}
+	}
+	return false
+}
+
 // Draw the current screen
 func (u *UI) Draw() error {
 	w, h := termbox.Size()
 	u.dirListHeight = h - 3
 
 	// Plot
-	err := termbox.Clear(termbox.ColorDefault, termbox.ColorDefault)
-	if err != nil {
-		return errors.Wrap(err, "failed to clear screen")
-	}
+	termbox.Clear(termbox.ColorWhite, termbox.ColorBlack)
 
 	// Header line
 	Linef(0, 0, w, termbox.ColorBlack, termbox.ColorWhite, ' ', "rclone ncdu %s - use the arrow keys to navigate, press ? for help", fs.Version)
@@ -316,6 +357,7 @@ func (u *UI) Draw() error {
 		if perBar == 0 {
 			perBar = 1
 		}
+		showEmptyDir := u.hasEmptyDir()
 		dirPos := u.dirPosMap[u.path]
 		for i, j := range u.sortPerm[dirPos.offset:] {
 			entry := u.entries[j]
@@ -323,31 +365,69 @@ func (u *UI) Draw() error {
 			if y >= h-1 {
 				break
 			}
+			attrs, err := u.d.AttrI(u.sortPerm[n])
+			_, isSelected := u.selectedEntries[entry.String()]
 			fg := termbox.ColorWhite
+			if attrs.EntriesHaveErrors {
+				fg = termbox.ColorYellow
+			}
+			if err != nil {
+				fg = termbox.ColorRed
+			}
+			const colorLightYellow = termbox.ColorYellow + 8
+			if isSelected {
+				fg = colorLightYellow
+			}
 			bg := termbox.ColorBlack
 			if n == dirPos.entry {
 				fg, bg = bg, fg
 			}
-			size, count, isDir, readable := u.d.AttrI(u.sortPerm[n])
 			mark := ' '
-			if isDir {
+			if attrs.IsDir {
 				mark = '/'
 			}
+			fileFlag := ' '
 			message := ""
-			if !readable {
+			if !attrs.Readable {
 				message = " [not read yet]"
 			}
+			if attrs.CountUnknownSize > 0 {
+				message = fmt.Sprintf(" [%d of %d files have unknown size, size may be underestimated]", attrs.CountUnknownSize, attrs.Count)
+				fileFlag = '~'
+			}
+			if attrs.EntriesHaveErrors {
+				message = " [some subdirectories could not be read, size may be underestimated]"
+				fileFlag = '.'
+			}
+			if err != nil {
+				message = fmt.Sprintf(" [%s]", err)
+				fileFlag = '!'
+			}
 			extras := ""
 			if u.showCounts {
-				if count > 0 {
-					extras += fmt.Sprintf("%8v ", fs.SizeSuffix(count))
+				ss := operations.CountStringField(attrs.Count, u.humanReadable, 9) + " "
+				if attrs.Count > 0 {
+					extras += ss
 				} else {
-					extras += "         "
+					extras += strings.Repeat(" ", len(ss))
+				}
+			}
+			if u.showDirAverageSize {
+				avg := attrs.AverageSize()
+				ss := operations.SizeStringField(int64(avg), u.humanReadable, 9) + " "
+				if avg > 0 {
+					extras += ss
+				} else {
+					extras += strings.Repeat(" ", len(ss))
+				}
+			}
+			if showEmptyDir {
+				if attrs.IsDir && attrs.Count == 0 && fileFlag == ' ' {
+					fileFlag = 'e'
 				}
-
 			}
 			if u.showGraph {
-				bars := (size + perBar/2 - 1) / perBar
+				bars := (attrs.Size + perBar/2 - 1) / perBar
 				// clip if necessary - only happens during startup
 				if bars > 10 {
 					bars = 10
@@ -356,7 +436,7 @@ func (u *UI) Draw() error {
 				}
 				extras += "[" + graph[graphBars-bars:2*graphBars-bars] + "] "
 			}
-			Linef(0, y, w, fg, bg, ' ', "%8v %s%c%s%s", fs.SizeSuffix(size), extras, mark, path.Base(entry.Remote()), message)
+			Linef(0, y, w, fg, bg, ' ', "%c %s %s%c%s%s", fileFlag, operations.SizeStringField(attrs.Size, u.humanReadable, 12), extras, mark, path.Base(entry.Remote()), message)
 			y++
 		}
 	}
@@ -370,16 +450,16 @@ func (u *UI) Draw() error {
 			message = " [listing in progress]"
 		}
 		size, count := u.d.Attr()
-		Linef(0, h-1, w, termbox.ColorBlack, termbox.ColorWhite, ' ', "Total usage: %v, Objects: %d%s", fs.SizeSuffix(size), count, message)
+		Linef(0, h-1, w, termbox.ColorBlack, termbox.ColorWhite, ' ', "Total usage: %s, Objects: %s%s", operations.SizeString(size, u.humanReadable), operations.CountString(count, u.humanReadable), message)
 	}
 
 	// Show the box on top if required
 	if u.showBox {
 		u.Box()
 	}
-	err = termbox.Flush()
+	err := termbox.Flush()
 	if err != nil {
-		return errors.Wrap(err, "failed to flush screen")
+		return fmt.Errorf("failed to flush screen: %w", err)
 	}
 	return nil
 }
@@ -424,6 +504,11 @@ func (u *UI) move(d int) {
 		dirPos.offset = entries - 1
 	}
 
+	// toggle the current file for selection in selection mode
+	if u.visualSelectMode {
+		u.toggleSelectForCursor()
+	}
+
 	// write dirPos back for later
 	u.dirPosMap[u.path] = dirPos
 }
@@ -433,13 +518,25 @@ func (u *UI) removeEntry(pos int) {
 	u.setCurrentDir(u.d)
 }
 
-// delete the entry at the current position
 func (u *UI) delete() {
+	if u.d == nil || len(u.entries) == 0 {
+		return
+	}
+	if len(u.selectedEntries) > 0 {
+		u.deleteSelected()
+	} else {
+		u.deleteSingle()
+	}
+}
+
+// delete the entry at the current position
+func (u *UI) deleteSingle() {
 	ctx := context.Background()
-	dirPos := u.sortPerm[u.dirPosMap[u.path].entry]
-	entry := u.entries[dirPos]
+	cursorPos := u.dirPosMap[u.path]
+	dirPos := u.sortPerm[cursorPos.entry]
+	dirEntry := u.entries[dirPos]
 	u.boxMenu = []string{"cancel", "confirm"}
-	if obj, isFile := entry.(fs.Object); isFile {
+	if obj, isFile := dirEntry.(fs.Object); isFile {
 		u.boxMenuHandler = func(f fs.Fs, p string, o int) (string, error) {
 			if o != 1 {
 				return "Aborted!", nil
@@ -449,30 +546,92 @@ func (u *UI) delete() {
 				return "", err
 			}
 			u.removeEntry(dirPos)
+			if cursorPos.entry >= len(u.entries) {
+				u.move(-1) // move back onto a valid entry
+			}
 			return "Successfully deleted file!", nil
 		}
 		u.popupBox([]string{
 			"Delete this file?",
-			u.fsName + entry.String()})
+			fspath.JoinRootPath(u.fsName, dirEntry.String())})
 	} else {
 		u.boxMenuHandler = func(f fs.Fs, p string, o int) (string, error) {
 			if o != 1 {
 				return "Aborted!", nil
 			}
-			err := operations.Purge(ctx, f, entry.String())
+			err := operations.Purge(ctx, f, dirEntry.String())
 			if err != nil {
 				return "", err
 			}
 			u.removeEntry(dirPos)
+			if cursorPos.entry >= len(u.entries) {
+				u.move(-1) // move back onto a valid entry
+			}
 			return "Successfully purged folder!", nil
 		}
 		u.popupBox([]string{
 			"Purge this directory?",
 			"ALL files in it will be deleted",
-			u.fsName + entry.String()})
+			fspath.JoinRootPath(u.fsName, dirEntry.String())})
 	}
 }
 
+func (u *UI) deleteSelected() {
+	ctx := context.Background()
+
+	u.boxMenu = []string{"cancel", "confirm"}
+
+	u.boxMenuHandler = func(f fs.Fs, p string, o int) (string, error) {
+		if o != 1 {
+			return "Aborted!", nil
+		}
+
+		positionsToDelete := make([]int, len(u.selectedEntries))
+		i := 0
+
+		for key, cursorPos := range u.selectedEntries {
+
+			dirPos := u.sortPerm[cursorPos.entry]
+			dirEntry := u.entries[dirPos]
+			var err error
+
+			if obj, isFile := dirEntry.(fs.Object); isFile {
+				err = operations.DeleteFile(ctx, obj)
+			} else {
+				err = operations.Purge(ctx, f, dirEntry.String())
+			}
+
+			if err != nil {
+				return "", err
+			}
+
+			delete(u.selectedEntries, key)
+			positionsToDelete[i] = dirPos
+			i++
+		}
+
+		// deleting all entries at once, as doing it during the deletions
+		// could cause issues.
+		sort.Slice(positionsToDelete, func(i, j int) bool {
+			return positionsToDelete[i] > positionsToDelete[j]
+		})
+		for _, dirPos := range positionsToDelete {
+			u.removeEntry(dirPos)
+		}
+
+		// move cursor at end if needed
+		cursorPos := u.dirPosMap[u.path]
+		if cursorPos.entry >= len(u.entries) {
+			u.move(-1)
+		}
+
+		return "Successfully deleted all items!", nil
+	}
+	u.popupBox([]string{
+		"Delete selected items?",
+		fmt.Sprintf("ALL %d items will be deleted", len(u.selectedEntries))})
+}
+
 func (u *UI) displayPath() {
 	u.togglePopupBox([]string{
 		"Current Path",
@@ -496,30 +655,50 @@ type ncduSort struct {
 
 // Less is part of sort.Interface.
 func (ds *ncduSort) Less(i, j int) bool {
-	isize, icount, _, _ := ds.d.AttrI(ds.sortPerm[i])
-	jsize, jcount, _, _ := ds.d.AttrI(ds.sortPerm[j])
+	var iAvgSize, jAvgSize float64
+	iattrs, _ := ds.d.AttrI(ds.sortPerm[i])
+	jattrs, _ := ds.d.AttrI(ds.sortPerm[j])
 	iname, jname := ds.entries[ds.sortPerm[i]].Remote(), ds.entries[ds.sortPerm[j]].Remote()
+	if iattrs.Count > 0 {
+		iAvgSize = iattrs.AverageSize()
+	}
+	if jattrs.Count > 0 {
+		jAvgSize = jattrs.AverageSize()
+	}
+
 	switch {
 	case ds.u.sortByName < 0:
 		return iname > jname
 	case ds.u.sortByName > 0:
 		break
 	case ds.u.sortBySize < 0:
-		if isize != jsize {
-			return isize < jsize
+		if iattrs.Size != jattrs.Size {
+			return iattrs.Size < jattrs.Size
 		}
 	case ds.u.sortBySize > 0:
-		if isize != jsize {
-			return isize > jsize
+		if iattrs.Size != jattrs.Size {
+			return iattrs.Size > jattrs.Size
 		}
 	case ds.u.sortByCount < 0:
-		if icount != jcount {
-			return icount < jcount
+		if iattrs.Count != jattrs.Count {
+			return iattrs.Count < jattrs.Count
 		}
 	case ds.u.sortByCount > 0:
-		if icount != jcount {
-			return icount > jcount
+		if iattrs.Count != jattrs.Count {
+			return iattrs.Count > jattrs.Count
 		}
+	case ds.u.sortByAverageSize < 0:
+		if iAvgSize != jAvgSize {
+			return iAvgSize < jAvgSize
+		}
+		// if avgSize is equal, sort by size
+		return iattrs.Size < jattrs.Size
+	case ds.u.sortByAverageSize > 0:
+		if iAvgSize != jAvgSize {
+			return iAvgSize > jAvgSize
+		}
+		// if avgSize is equal, sort by size
+		return iattrs.Size > jattrs.Size
 	}
 	// if everything equal, sort by name
 	return iname < jname
@@ -560,7 +739,9 @@ func (u *UI) sortCurrentDir() {
 func (u *UI) setCurrentDir(d *scan.Dir) {
 	u.d = d
 	u.entries = d.Entries()
-	u.path = path.Join(u.fsName, d.Path())
+	u.path = fspath.JoinRootPath(u.fsName, d.Path())
+	u.selectedEntries = make(map[string]dirPos)
+	u.visualSelectMode = false
 	u.sortCurrentDir()
 }
 
@@ -628,6 +809,7 @@ func (u *UI) toggleSort(sortType *int8) {
 	u.sortBySize = 0
 	u.sortByCount = 0
 	u.sortByName = 0
+	u.sortByAverageSize = 0
 	if old == 0 {
 		*sortType = 1
 	} else {
@@ -636,19 +818,36 @@ func (u *UI) toggleSort(sortType *int8) {
 	u.sortCurrentDir()
 }
 
+func (u *UI) toggleSelectForCursor() {
+	cursorPos := u.dirPosMap[u.path]
+	dirPos := u.sortPerm[cursorPos.entry]
+	dirEntry := u.entries[dirPos]
+
+	_, present := u.selectedEntries[dirEntry.String()]
+
+	if present {
+		delete(u.selectedEntries, dirEntry.String())
+	} else {
+		u.selectedEntries[dirEntry.String()] = cursorPos
+	}
+}
+
 // NewUI creates a new user interface for ncdu on f
 func NewUI(f fs.Fs) *UI {
 	return &UI{
-		f:             f,
-		path:          "Waiting for root...",
-		dirListHeight: 20, // updated in Draw
-		fsName:        f.Name() + ":" + f.Root(),
-		showGraph:     true,
-		showCounts:    false,
-		sortByName:    0, // +1 for normal, 0 for off, -1 for reverse
-		sortBySize:    1,
-		sortByCount:   0,
-		dirPosMap:     make(map[string]dirPos),
+		f:                  f,
+		path:               "Waiting for root...",
+		dirListHeight:      20, // updated in Draw
+		fsName:             fs.ConfigString(f),
+		showGraph:          true,
+		showCounts:         false,
+		showDirAverageSize: false,
+		humanReadable:      true,
+		sortByName:         0, // +1 for normal, 0 for off, -1 for reverse
+		sortBySize:         1,
+		sortByCount:        0,
+		dirPosMap:          make(map[string]dirPos),
+		selectedEntries:    make(map[string]dirPos),
 	}
 }
 
@@ -656,7 +855,7 @@ func NewUI(f fs.Fs) *UI {
 func (u *UI) Show() error {
 	err := termbox.Init()
 	if err != nil {
-		return errors.Wrap(err, "termbox init")
+		return fmt.Errorf("termbox init: %w", err)
 	}
 	defer termbox.Close()
 
@@ -680,7 +879,7 @@ outer:
 		//Reset()
 		err := u.Draw()
 		if err != nil {
-			return errors.Wrap(err, "draw failed")
+			return fmt.Errorf("draw failed: %w", err)
 		}
 		var root *scan.Dir
 		select {
@@ -689,7 +888,7 @@ outer:
 			u.setCurrentDir(root)
 		case err := <-errChan:
 			if err != nil {
-				return errors.Wrap(err, "ncdu directory listing")
+				return fmt.Errorf("ncdu directory listing: %w", err)
 			}
 			u.listing = false
 		case <-updated:
@@ -736,18 +935,30 @@ outer:
 					u.showCounts = !u.showCounts
 				case 'g':
 					u.showGraph = !u.showGraph
+				case 'a':
+					u.showDirAverageSize = !u.showDirAverageSize
 				case 'n':
 					u.toggleSort(&u.sortByName)
 				case 's':
 					u.toggleSort(&u.sortBySize)
+				case 'v':
+					u.toggleSelectForCursor()
+				case 'V':
+					u.visualSelectMode = !u.visualSelectMode
 				case 'C':
 					u.toggleSort(&u.sortByCount)
+				case 'A':
+					u.toggleSort(&u.sortByAverageSize)
 				case 'y':
 					u.copyPath()
 				case 'Y':
 					u.displayPath()
 				case 'd':
 					u.delete()
+				case 'u':
+					u.humanReadable = !u.humanReadable
+				case 'D':
+					u.deleteSelected()
 				case '?':
 					u.togglePopupBox(helpText())
 
@@ -761,7 +972,7 @@ outer:
 				}
 			}
 		}
-		// listen to key presses, etc
+		// listen to key presses, etc.
 	}
 	return nil
 }
