@@ -2,23 +2,54 @@
 package scan
 
 import (
+	"context"
+	"fmt"
 	"path"
 	"sync"
 
-	"github.com/ncw/rclone/fs"
-	"github.com/ncw/rclone/fs/walk"
-	"github.com/pkg/errors"
+	"github.com/rclone/rclone/fs"
+	"github.com/rclone/rclone/fs/walk"
 )
 
 // Dir represents a directory found in the remote
 type Dir struct {
-	parent  *Dir
-	path    string
-	mu      sync.Mutex
-	count   int64
-	size    int64
-	entries fs.DirEntries
-	dirs    map[string]*Dir
+	parent            *Dir
+	path              string
+	mu                sync.Mutex
+	size              int64
+	count             int64
+	countUnknownSize  int64
+	entries           fs.DirEntries
+	dirs              map[string]*Dir
+	readError         error
+	entriesHaveErrors bool
+}
+
+// Attrs contains accumulated properties for a directory entry
+//
+// Files with unknown size are counted separately but also included
+// in the total count. They are not included in the size, i.e. treated
+// as empty files, which means the size may be underestimated.
+type Attrs struct {
+	Size              int64
+	Count             int64
+	CountUnknownSize  int64
+	IsDir             bool
+	Readable          bool
+	EntriesHaveErrors bool
+}
+
+// AverageSize calculates average size of files in directory
+//
+// If there are files with unknown size, this returns the average over
+// files with known sizes, which means it may be under- or
+// overestimated.
+func (a *Attrs) AverageSize() float64 {
+	countKnownSize := a.Count - a.CountUnknownSize
+	if countKnownSize > 0 {
+		return float64(a.Size) / float64(countKnownSize)
+	}
+	return 0
 }
 
 // Parent returns the directory above this one
@@ -34,18 +65,25 @@ func (d *Dir) Path() string {
 }
 
 // make a new directory
-func newDir(parent *Dir, dirPath string, entries fs.DirEntries) *Dir {
+func newDir(parent *Dir, dirPath string, entries fs.DirEntries, err error) *Dir {
 	d := &Dir{
-		parent:  parent,
-		path:    dirPath,
-		entries: entries,
-		dirs:    make(map[string]*Dir),
+		parent:    parent,
+		path:      dirPath,
+		entries:   entries,
+		dirs:      make(map[string]*Dir),
+		readError: err,
 	}
 	// Count size in this dir
 	for _, entry := range entries {
 		if o, ok := entry.(fs.Object); ok {
 			d.count++
-			d.size += o.Size()
+			size := o.Size()
+			if size < 0 {
+				// Some backends may return -1 because size of object is not known
+				d.countUnknownSize++
+			} else {
+				d.size += size
+			}
 		}
 	}
 	// Set my directory entry in parent
@@ -58,8 +96,12 @@ func newDir(parent *Dir, dirPath string, entries fs.DirEntries) *Dir {
 	// Accumulate counts in parents
 	for ; parent != nil; parent = parent.parent {
 		parent.mu.Lock()
-		parent.count += d.count
 		parent.size += d.size
+		parent.count += d.count
+		parent.countUnknownSize += d.countUnknownSize
+		if d.readError != nil {
+			parent.entriesHaveErrors = true
+		}
 		parent.mu.Unlock()
 	}
 	return d
@@ -84,17 +126,24 @@ func (d *Dir) Remove(i int) {
 // Call with d.mu held
 func (d *Dir) remove(i int) {
 	size := d.entries[i].Size()
+	countUnknownSize := int64(0)
+	if size < 0 {
+		size = 0
+		countUnknownSize = 1
+	}
 	count := int64(1)
 
 	subDir, ok := d.getDir(i)
 	if ok {
 		size = subDir.size
 		count = subDir.count
+		countUnknownSize = subDir.countUnknownSize
 		delete(d.dirs, path.Base(subDir.path))
 	}
 
 	d.size -= size
 	d.count -= count
+	d.countUnknownSize -= countUnknownSize
 	d.entries = append(d.entries[:i], d.entries[i+1:]...)
 
 	dir := d
@@ -104,6 +153,7 @@ func (d *Dir) remove(i int) {
 		parent.dirs[path.Base(dir.path)] = dir
 		parent.size -= size
 		parent.count -= count
+		parent.countUnknownSize -= countUnknownSize
 		dir = parent
 		parent.mu.Unlock()
 	}
@@ -144,32 +194,31 @@ func (d *Dir) Attr() (size int64, count int64) {
 }
 
 // AttrI returns the size, count and flags for the i-th directory entry
-func (d *Dir) AttrI(i int) (size int64, count int64, isDir bool, readable bool) {
+func (d *Dir) AttrI(i int) (attrs Attrs, err error) {
 	d.mu.Lock()
 	defer d.mu.Unlock()
 	subDir, isDir := d.getDir(i)
+
 	if !isDir {
-		return d.entries[i].Size(), 0, false, true
+		return Attrs{d.entries[i].Size(), 0, 0, false, true, d.entriesHaveErrors}, d.readError
 	}
 	if subDir == nil {
-		return 0, 0, true, false
+		return Attrs{0, 0, 0, true, false, false}, nil
 	}
-	size, count = subDir.Attr()
-	return size, count, true, true
+	size, count := subDir.Attr()
+	return Attrs{size, count, subDir.countUnknownSize, true, true, subDir.entriesHaveErrors}, subDir.readError
 }
 
 // Scan the Fs passed in, returning a root directory channel and an
 // error channel
-func Scan(f fs.Fs) (chan *Dir, chan error, chan struct{}) {
+func Scan(ctx context.Context, f fs.Fs) (chan *Dir, chan error, chan struct{}) {
+	ci := fs.GetConfig(ctx)
 	root := make(chan *Dir, 1)
 	errChan := make(chan error, 1)
 	updated := make(chan struct{}, 1)
 	go func() {
 		parents := map[string]*Dir{}
-		err := walk.Walk(f, "", false, fs.Config.MaxDepth, func(dirPath string, entries fs.DirEntries, err error) error {
-			if err != nil {
-				return err // FIXME mark directory as errored instead of aborting
-			}
+		err := walk.Walk(ctx, f, "", false, ci.MaxDepth, func(dirPath string, entries fs.DirEntries, err error) error {
 			var parent *Dir
 			if dirPath != "" {
 				parentPath := path.Dir(dirPath)
@@ -179,10 +228,10 @@ func Scan(f fs.Fs) (chan *Dir, chan error, chan struct{}) {
 				var ok bool
 				parent, ok = parents[parentPath]
 				if !ok {
-					errChan <- errors.Errorf("couldn't find parent for %q", dirPath)
+					errChan <- fmt.Errorf("couldn't find parent for %q", dirPath)
 				}
 			}
-			d := newDir(parent, dirPath, entries)
+			d := newDir(parent, dirPath, entries, err)
 			parents[dirPath] = d
 			if dirPath == "" {
 				root <- d
@@ -196,7 +245,7 @@ func Scan(f fs.Fs) (chan *Dir, chan error, chan struct{}) {
 			return nil
 		})
 		if err != nil {
-			errChan <- errors.Wrap(err, "ncdu listing failed")
+			errChan <- fmt.Errorf("ncdu listing failed: %w", err)
 		}
 		errChan <- nil
 	}()
