@@ -1,3 +1,4 @@
+// Package rmdir provides the rmdir command.
 package rmdir
 
 import (
@@ -20,15 +21,22 @@ func init() {
 var rmdirsCmd = &cobra.Command{
 	Use:   "rmdirs remote:path",
 	Short: `Remove empty directories under the path.`,
-	Long: `This removes any empty directories (or directories that only contain
-empty directories) under the path that it finds, including the path if
-it has nothing in.
+	Long: `
+This recursively removes any empty directories (including directories
+that only contain empty directories), that it finds under the path.
+The root path itself will also be removed if it is empty, unless
+you supply the ` + "`--leave-root`" + ` flag.
 
-If you supply the --leave-root flag, it will not remove the root directory.
+Use command [rmdir](/commands/rclone_rmdir/) to delete just the empty
+directory given by path, not recurse.
 
 This is useful for tidying up remotes that rclone has left a lot of
-empty directories in.
+empty directories in. For example the [delete](/commands/rclone_delete/)
+command will delete files but leave the directory structure (unless
+used with option ` + "`--rmdirs`" + `).
 
+To delete a path and any objects in it, use [purge](/commands/rclone_purge/)
+command.
 `,
 	Run: func(command *cobra.Command, args []string) {
 		cmd.CheckArgs(1, 1, command, args)
