@@ -1,6 +1,7 @@
 // Accounting and limiting reader
 // Unix specific functions.
 
+//go:build darwin || dragonfly || freebsd || linux || netbsd || openbsd || solaris
 // +build darwin dragonfly freebsd linux netbsd openbsd solaris
 
 package accounting
@@ -10,11 +11,11 @@ import (
 	"os/signal"
 	"syscall"
 
-	"github.com/ncw/rclone/fs"
+	"github.com/rclone/rclone/fs"
 )
 
 // startSignalHandler() sets a signal handler to catch SIGUSR2 and toggle throttling.
-func startSignalHandler() {
+func (tb *tokenBucket) startSignalHandler() {
 	signals := make(chan os.Signal, 1)
 	signal.Notify(signals, syscall.SIGUSR2)
 
@@ -22,14 +23,14 @@ func startSignalHandler() {
 		// This runs forever, but blocks until the signal is received.
 		for {
 			<-signals
-			tokenBucketMu.Lock()
-			bwLimitToggledOff = !bwLimitToggledOff
-			tokenBucket, prevTokenBucket = prevTokenBucket, tokenBucket
+			tb.mu.Lock()
+			tb.toggledOff = !tb.toggledOff
+			tb.curr, tb.prev = tb.prev, tb.curr
 			s := "disabled"
-			if tokenBucket != nil {
+			if !tb.curr._isOff() {
 				s = "enabled"
 			}
-			tokenBucketMu.Unlock()
+			tb.mu.Unlock()
 			fs.Logf(nil, "Bandwidth limit %s by user", s)
 		}
 	}()
