@@ -1,3 +1,4 @@
+// Package chunkedreader provides functionality for reading in chunks.
 package chunkedreader
 
 import (
@@ -7,6 +8,7 @@ import (
 	"sync"
 
 	"github.com/rclone/rclone/fs"
+	"github.com/rclone/rclone/fs/hash"
 )
 
 // io related errors returned by ChunkedReader
@@ -15,10 +17,10 @@ var (
 	ErrorInvalidSeek = errors.New("invalid seek position")
 )
 
-// ChunkedReader is a reader for a Object with the possibility
+// ChunkedReader is a reader for an Object with the possibility
 // of reading the source in chunks of given size
 //
-// A initialChunkSize of <= 0 will disable chunked reading.
+// An initialChunkSize of <= 0 will disable chunked reading.
 type ChunkedReader struct {
 	ctx              context.Context
 	mu               sync.Mutex    // protects following fields
@@ -35,9 +37,9 @@ type ChunkedReader struct {
 
 // New returns a ChunkedReader for the Object.
 //
-// A initialChunkSize of <= 0 will disable chunked reading.
+// An initialChunkSize of <= 0 will disable chunked reading.
 // If maxChunkSize is greater than initialChunkSize, the chunk size will be
-// doubled after each chunk read with a maximun of maxChunkSize.
+// doubled after each chunk read with a maximum of maxChunkSize.
 // A Seek or RangeSeek will reset the chunk size to it's initial value
 func New(ctx context.Context, o fs.Object, initialChunkSize int64, maxChunkSize int64) *ChunkedReader {
 	if initialChunkSize <= 0 {
@@ -215,12 +217,12 @@ func (cr *ChunkedReader) openRange() error {
 	var err error
 	if length <= 0 {
 		if offset == 0 {
-			rc, err = cr.o.Open(cr.ctx)
+			rc, err = cr.o.Open(cr.ctx, &fs.HashesOption{Hashes: hash.Set(hash.None)})
 		} else {
-			rc, err = cr.o.Open(cr.ctx, &fs.RangeOption{Start: offset, End: -1})
+			rc, err = cr.o.Open(cr.ctx, &fs.HashesOption{Hashes: hash.Set(hash.None)}, &fs.RangeOption{Start: offset, End: -1})
 		}
 	} else {
-		rc, err = cr.o.Open(cr.ctx, &fs.RangeOption{Start: offset, End: offset + length - 1})
+		rc, err = cr.o.Open(cr.ctx, &fs.HashesOption{Hashes: hash.Set(hash.None)}, &fs.RangeOption{Start: offset, End: offset + length - 1})
 	}
 	if err != nil {
 		return err
