@@ -1,5 +1,6 @@
 // Log the panic to the log file - for oses which can't do this
 
+//go:build !windows && !darwin && !dragonfly && !freebsd && !linux && !nacl && !netbsd && !openbsd
 // +build !windows,!darwin,!dragonfly,!freebsd,!linux,!nacl,!netbsd,!openbsd
 
 package log
@@ -7,7 +8,7 @@ package log
 import (
 	"os"
 
-	"github.com/ncw/rclone/fs"
+	"github.com/rclone/rclone/fs"
 )
 
 // redirectStderr to the file passed in
