@@ -2,10 +2,12 @@
 package rcserver
 
 import (
+	"context"
 	"encoding/base64"
 	"encoding/json"
 	"flag"
 	"fmt"
+	"log"
 	"mime"
 	"net/http"
 	"net/url"
@@ -13,28 +15,51 @@ import (
 	"regexp"
 	"sort"
 	"strings"
+	"sync"
+	"time"
 
-	"github.com/pkg/errors"
+	"github.com/prometheus/client_golang/prometheus"
+	"github.com/prometheus/client_golang/prometheus/promhttp"
 	"github.com/rclone/rclone/cmd/serve/httplib"
-	"github.com/rclone/rclone/cmd/serve/httplib/serve"
 	"github.com/rclone/rclone/fs"
+	"github.com/rclone/rclone/fs/accounting"
 	"github.com/rclone/rclone/fs/cache"
 	"github.com/rclone/rclone/fs/config"
+	"github.com/rclone/rclone/fs/fshttp"
 	"github.com/rclone/rclone/fs/list"
 	"github.com/rclone/rclone/fs/rc"
 	"github.com/rclone/rclone/fs/rc/jobs"
 	"github.com/rclone/rclone/fs/rc/rcflags"
+	"github.com/rclone/rclone/fs/rc/webgui"
+	"github.com/rclone/rclone/lib/http/serve"
+	"github.com/rclone/rclone/lib/random"
 	"github.com/skratchdot/open-golang/open"
 )
 
+var promHandler http.Handler
+var onlyOnceWarningAllowOrigin sync.Once
+
+func init() {
+	rcloneCollector := accounting.NewRcloneCollector(context.Background())
+	prometheus.MustRegister(rcloneCollector)
+
+	m := fshttp.NewMetrics("rclone")
+	for _, c := range m.Collectors() {
+		prometheus.MustRegister(c)
+	}
+	fshttp.DefaultMetrics = m
+
+	promHandler = promhttp.Handler()
+}
+
 // Start the remote control server if configured
 //
 // If the server wasn't configured the *Server returned may be nil
-func Start(opt *rc.Options) (*Server, error) {
+func Start(ctx context.Context, opt *rc.Options) (*Server, error) {
 	jobs.SetOpt(opt) // set the defaults for jobs
 	if opt.Enabled {
 		// Serve on the DefaultServeMux so can have global registrations appear
-		s := newServer(opt, http.DefaultServeMux)
+		s := newServer(ctx, opt, http.DefaultServeMux)
 		return s, s.Serve()
 	}
 	return nil, nil
@@ -43,22 +68,20 @@ func Start(opt *rc.Options) (*Server, error) {
 // Server contains everything to run the rc server
 type Server struct {
 	*httplib.Server
-	files http.Handler
-	opt   *rc.Options
+	ctx            context.Context // for global config
+	files          http.Handler
+	pluginsHandler http.Handler
+	opt            *rc.Options
 }
 
-func newServer(opt *rc.Options, mux *http.ServeMux) *Server {
-	s := &Server{
-		Server: httplib.NewServer(mux, &opt.HTTPOptions),
-		opt:    opt,
-	}
-	mux.HandleFunc("/", s.handler)
-
+func newServer(ctx context.Context, opt *rc.Options, mux *http.ServeMux) *Server {
+	fileHandler := http.Handler(nil)
+	pluginsHandler := http.Handler(nil)
 	// Add some more mime types which are often missing
 	_ = mime.AddExtensionType(".wasm", "application/wasm")
 	_ = mime.AddExtensionType(".js", "application/javascript")
 
-	cachePath := filepath.Join(config.CacheDir, "webgui")
+	cachePath := filepath.Join(config.GetCacheDir(), "webgui")
 	extractPath := filepath.Join(cachePath, "current/build")
 	// File handling
 	if opt.Files != "" {
@@ -66,10 +89,44 @@ func newServer(opt *rc.Options, mux *http.ServeMux) *Server {
 			fs.Logf(nil, "--rc-files overrides --rc-web-gui command\n")
 		}
 		fs.Logf(nil, "Serving files from %q", opt.Files)
-		s.files = http.FileServer(http.Dir(opt.Files))
+		fileHandler = http.FileServer(http.Dir(opt.Files))
 	} else if opt.WebUI {
-		s.files = http.FileServer(http.Dir(extractPath))
+		if err := webgui.CheckAndDownloadWebGUIRelease(opt.WebGUIUpdate, opt.WebGUIForceUpdate, opt.WebGUIFetchURL, config.GetCacheDir()); err != nil {
+			fs.Errorf(nil, "Error while fetching the latest release of Web GUI: %v", err)
+		}
+		if opt.NoAuth {
+			fs.Logf(nil, "It is recommended to use web gui with auth.")
+		} else {
+			if opt.HTTPOptions.BasicUser == "" {
+				opt.HTTPOptions.BasicUser = "gui"
+				fs.Infof(nil, "No username specified. Using default username: %s \n", rcflags.Opt.HTTPOptions.BasicUser)
+			}
+			if opt.HTTPOptions.BasicPass == "" {
+				randomPass, err := random.Password(128)
+				if err != nil {
+					log.Fatalf("Failed to make password: %v", err)
+				}
+				opt.HTTPOptions.BasicPass = randomPass
+				fs.Infof(nil, "No password specified. Using random password: %s \n", randomPass)
+			}
+		}
+		opt.Serve = true
+
+		fs.Logf(nil, "Serving Web GUI")
+		fileHandler = http.FileServer(http.Dir(extractPath))
+
+		pluginsHandler = http.FileServer(http.Dir(webgui.PluginsPath))
 	}
+
+	s := &Server{
+		Server:         httplib.NewServer(mux, &opt.HTTPOptions),
+		ctx:            ctx,
+		opt:            opt,
+		files:          fileHandler,
+		pluginsHandler: pluginsHandler,
+	}
+	mux.HandleFunc("/", s.handler)
+
 	return s
 }
 
@@ -86,7 +143,7 @@ func (s *Server) Serve() error {
 	if s.files != nil {
 		openURL, err := url.Parse(s.URL())
 		if err != nil {
-			return errors.Wrap(err, "invalid serving URL")
+			return fmt.Errorf("invalid serving URL: %w", err)
 		}
 		// Add username, password into the URL if they are set
 		user, pass := s.opt.HTTPOptions.BasicUser, s.opt.HTTPOptions.BasicPass
@@ -102,11 +159,13 @@ func (s *Server) Serve() error {
 			openURL.RawQuery = parameters.Encode()
 			openURL.RawPath = "/#/login"
 		}
-		// Don't open browser if serving in testing environment.
-		if flag.Lookup("test.v") == nil {
-			_ = open.Start(openURL.String())
+		// Don't open browser if serving in testing environment or required not to do so.
+		if flag.Lookup("test.v") == nil && !s.opt.WebGUINoOpenBrowser {
+			if err := open.Start(openURL.String()); err != nil {
+				fs.Errorf(nil, "Failed to open Web GUI in browser: %v. Manually access it at: %s", err, openURL.String())
+			}
 		} else {
-			fs.Errorf(nil, "Not opening browser in testing environment")
+			fs.Logf(nil, "Web GUI is not automatically opening browser. Navigate to %s to use.", openURL.String())
 		}
 	}
 	return nil
@@ -115,24 +174,12 @@ func (s *Server) Serve() error {
 // writeError writes a formatted error to the output
 func writeError(path string, in rc.Params, w http.ResponseWriter, err error, status int) {
 	fs.Errorf(nil, "rc: %q: error: %v", path, err)
-	// Adjust the error return for some well known errors
-	errOrig := errors.Cause(err)
-	switch {
-	case errOrig == fs.ErrorDirNotFound || errOrig == fs.ErrorObjectNotFound:
-		status = http.StatusNotFound
-	case rc.IsErrParamInvalid(err) || rc.IsErrParamNotFound(err):
-		status = http.StatusBadRequest
-	}
+	params, status := rc.Error(path, in, err, status)
 	w.WriteHeader(status)
-	err = rc.WriteJSON(w, rc.Params{
-		"status": status,
-		"error":  err.Error(),
-		"input":  in,
-		"path":   path,
-	})
+	err = rc.WriteJSON(w, params)
 	if err != nil {
 		// can't return the error at this point
-		fs.Errorf(nil, "rc: failed to write JSON output: %v", err)
+		fs.Errorf(nil, "rc: writeError: failed to write JSON output from %#v: %v", in, err)
 	}
 }
 
@@ -146,9 +193,11 @@ func (s *Server) handler(w http.ResponseWriter, r *http.Request) {
 
 	allowOrigin := rcflags.Opt.AccessControlAllowOrigin
 	if allowOrigin != "" {
-		if allowOrigin == "*" {
-			fs.Logf(nil, "Warning: Allow origin set to *. This can cause serious security problems.")
-		}
+		onlyOnceWarningAllowOrigin.Do(func() {
+			if allowOrigin == "*" {
+				fs.Logf(nil, "Warning: Allow origin set to *. This can cause serious security problems.")
+			}
+		})
 		w.Header().Add("Access-Control-Allow-Origin", allowOrigin)
 	} else {
 		w.Header().Add("Access-Control-Allow-Origin", s.URL())
@@ -167,12 +216,13 @@ func (s *Server) handler(w http.ResponseWriter, r *http.Request) {
 	case "GET", "HEAD":
 		s.handleGet(w, r, path)
 	default:
-		writeError(path, nil, w, errors.Errorf("method %q not allowed", r.Method), http.StatusMethodNotAllowed)
+		writeError(path, nil, w, fmt.Errorf("method %q not allowed", r.Method), http.StatusMethodNotAllowed)
 		return
 	}
 }
 
 func (s *Server) handlePost(w http.ResponseWriter, r *http.Request, path string) {
+	ctx := r.Context()
 	contentType := r.Header.Get("Content-Type")
 
 	values := r.URL.Query()
@@ -180,7 +230,7 @@ func (s *Server) handlePost(w http.ResponseWriter, r *http.Request, path string)
 		// Parse the POST and URL parameters into r.Form, for others r.Form will be empty value
 		err := r.ParseForm()
 		if err != nil {
-			writeError(path, nil, w, errors.Wrap(err, "failed to parse form/URL parameters"), http.StatusBadRequest)
+			writeError(path, nil, w, fmt.Errorf("failed to parse form/URL parameters: %w", err), http.StatusBadRequest)
 			return
 		}
 		values = r.Form
@@ -198,43 +248,41 @@ func (s *Server) handlePost(w http.ResponseWriter, r *http.Request, path string)
 	if contentType == "application/json" {
 		err := json.NewDecoder(r.Body).Decode(&in)
 		if err != nil {
-			writeError(path, in, w, errors.Wrap(err, "failed to read input JSON"), http.StatusBadRequest)
+			writeError(path, in, w, fmt.Errorf("failed to read input JSON: %w", err), http.StatusBadRequest)
 			return
 		}
 	}
-
 	// Find the call
 	call := rc.Calls.Get(path)
 	if call == nil {
-		writeError(path, in, w, errors.Errorf("couldn't find method %q", path), http.StatusNotFound)
+		writeError(path, in, w, fmt.Errorf("couldn't find method %q", path), http.StatusNotFound)
 		return
 	}
 
 	// Check to see if it requires authorisation
 	if !s.opt.NoAuth && call.AuthRequired && !s.UsingAuth() {
-		writeError(path, in, w, errors.Errorf("authentication must be set up on the rc server to use %q or the --rc-no-auth flag must be in use", path), http.StatusForbidden)
+		writeError(path, in, w, fmt.Errorf("authentication must be set up on the rc server to use %q or the --rc-no-auth flag must be in use", path), http.StatusForbidden)
 		return
 	}
 
-	// Check to see if it is async or not
-	isAsync, err := in.GetBool("_async")
-	if rc.NotErrParamNotFound(err) {
-		writeError(path, in, w, err, http.StatusBadRequest)
-		return
+	inOrig := in.Copy()
+
+	if call.NeedsRequest {
+		// Add the request to RC
+		in["_request"] = r
+	}
+
+	if call.NeedsResponse {
+		in["_response"] = w
 	}
-	delete(in, "_async") // remove the async parameter after parsing so vfs operations don't get confused
 
 	fs.Debugf(nil, "rc: %q: with parameters %+v", path, in)
-	var out rc.Params
-	if isAsync {
-		out, err = jobs.StartAsyncJob(call.Fn, in)
-	} else {
-		var jobID int64
-		out, jobID, err = jobs.ExecuteJob(r.Context(), call.Fn, in)
-		w.Header().Add("x-rclone-jobid", fmt.Sprintf("%d", jobID))
+	job, out, err := jobs.NewJob(ctx, call.Fn, in)
+	if job != nil {
+		w.Header().Add("x-rclone-jobid", fmt.Sprintf("%d", job.ID))
 	}
 	if err != nil {
-		writeError(path, in, w, err, http.StatusInternalServerError)
+		writeError(path, inOrig, w, err, http.StatusInternalServerError)
 		return
 	}
 	if out == nil {
@@ -245,8 +293,8 @@ func (s *Server) handlePost(w http.ResponseWriter, r *http.Request, path string)
 	err = rc.WriteJSON(w, out)
 	if err != nil {
 		// can't return the error at this point - but have a go anyway
-		writeError(path, in, w, err, http.StatusInternalServerError)
-		fs.Errorf(nil, "rc: failed to write JSON output: %v", err)
+		writeError(path, inOrig, w, err, http.StatusInternalServerError)
+		fs.Errorf(nil, "rc: handlePost: failed to write JSON output: %v", err)
 	}
 }
 
@@ -258,40 +306,49 @@ func (s *Server) serveRoot(w http.ResponseWriter, r *http.Request) {
 	remotes := config.FileSections()
 	sort.Strings(remotes)
 	directory := serve.NewDirectory("", s.HTMLTemplate)
-	directory.Title = "List of all rclone remotes."
+	directory.Name = "List of all rclone remotes."
 	q := url.Values{}
 	for _, remote := range remotes {
 		q.Set("fs", remote)
-		directory.AddEntry("["+remote+":]", true)
+		directory.AddHTMLEntry("["+remote+":]", true, -1, time.Time{})
 	}
+	sortParm := r.URL.Query().Get("sort")
+	orderParm := r.URL.Query().Get("order")
+	directory.ProcessQueryParams(sortParm, orderParm)
+
 	directory.Serve(w, r)
 }
 
 func (s *Server) serveRemote(w http.ResponseWriter, r *http.Request, path string, fsName string) {
-	f, err := cache.Get(fsName)
+	f, err := cache.Get(s.ctx, fsName)
 	if err != nil {
-		writeError(path, nil, w, errors.Wrap(err, "failed to make Fs"), http.StatusInternalServerError)
+		writeError(path, nil, w, fmt.Errorf("failed to make Fs: %w", err), http.StatusInternalServerError)
 		return
 	}
 	if path == "" || strings.HasSuffix(path, "/") {
 		path = strings.Trim(path, "/")
 		entries, err := list.DirSorted(r.Context(), f, false, path)
 		if err != nil {
-			writeError(path, nil, w, errors.Wrap(err, "failed to list directory"), http.StatusInternalServerError)
+			writeError(path, nil, w, fmt.Errorf("failed to list directory: %w", err), http.StatusInternalServerError)
 			return
 		}
 		// Make the entries for display
 		directory := serve.NewDirectory(path, s.HTMLTemplate)
 		for _, entry := range entries {
 			_, isDir := entry.(fs.Directory)
-			directory.AddEntry(entry.Remote(), isDir)
+			//directory.AddHTMLEntry(entry.Remote(), isDir, entry.Size(), entry.ModTime(r.Context()))
+			directory.AddHTMLEntry(entry.Remote(), isDir, entry.Size(), time.Time{})
 		}
+		sortParm := r.URL.Query().Get("sort")
+		orderParm := r.URL.Query().Get("order")
+		directory.ProcessQueryParams(sortParm, orderParm)
+
 		directory.Serve(w, r)
 	} else {
 		path = strings.Trim(path, "/")
 		o, err := f.NewObject(r.Context(), path)
 		if err != nil {
-			writeError(path, nil, w, errors.Wrap(err, "failed to find object"), http.StatusInternalServerError)
+			writeError(path, nil, w, fmt.Errorf("failed to find object: %w", err), http.StatusInternalServerError)
 			return
 		}
 		serve.Object(w, r, o)
@@ -303,17 +360,36 @@ var fsMatch = regexp.MustCompile(`^\[(.*?)\](.*)$`)
 
 func (s *Server) handleGet(w http.ResponseWriter, r *http.Request, path string) {
 	// Look to see if this has an fs in the path
-	match := fsMatch.FindStringSubmatch(path)
+	fsMatchResult := fsMatch.FindStringSubmatch(path)
+
 	switch {
-	case match != nil && s.opt.Serve:
+	case fsMatchResult != nil && s.opt.Serve:
 		// Serve /[fs]/remote files
-		s.serveRemote(w, r, match[2], match[1])
+		s.serveRemote(w, r, fsMatchResult[2], fsMatchResult[1])
+		return
+	case path == "metrics" && s.opt.EnableMetrics:
+		promHandler.ServeHTTP(w, r)
 		return
 	case path == "*" && s.opt.Serve:
 		// Serve /* as the remote listing
 		s.serveRoot(w, r)
 		return
 	case s.files != nil:
+		if s.opt.WebUI {
+			pluginsMatchResult := webgui.PluginsMatch.FindStringSubmatch(path)
+
+			if len(pluginsMatchResult) > 2 {
+				ok := webgui.ServePluginOK(w, r, pluginsMatchResult)
+				if !ok {
+					r.URL.Path = fmt.Sprintf("/%s/%s/app/build/%s", pluginsMatchResult[1], pluginsMatchResult[2], pluginsMatchResult[3])
+					s.pluginsHandler.ServeHTTP(w, r)
+					return
+				}
+				return
+			} else if webgui.ServePluginWithReferrerOK(w, r, path) {
+				return
+			}
+		}
 		// Serve the files
 		r.URL.Path = "/" + path
 		s.files.ServeHTTP(w, r)
