@@ -17,7 +17,7 @@ func TestMkdir(t *testing.T) {
 	// test stuff
 }
 
-This will make r.Fremote and r.Flocal for a remote remote and a local
+This will make r.Fremote and r.Flocal for a remote and a local
 remote.  The remote is determined by the -remote flag passed in.
 
 */
@@ -26,6 +26,7 @@ package fstest
 
 import (
 	"bytes"
+	"context"
 	"flag"
 	"fmt"
 	"io/ioutil"
@@ -37,10 +38,13 @@ import (
 	"testing"
 	"time"
 
-	"github.com/ncw/rclone/fs"
-	"github.com/ncw/rclone/fs/fserrors"
-	"github.com/ncw/rclone/fs/object"
-	"github.com/ncw/rclone/fs/walk"
+	"github.com/rclone/rclone/fs"
+	"github.com/rclone/rclone/fs/cache"
+	"github.com/rclone/rclone/fs/fserrors"
+	"github.com/rclone/rclone/fs/hash"
+	"github.com/rclone/rclone/fs/object"
+	"github.com/rclone/rclone/fs/walk"
+	"github.com/rclone/rclone/lib/file"
 	"github.com/stretchr/testify/assert"
 	"github.com/stretchr/testify/require"
 )
@@ -51,6 +55,7 @@ type Run struct {
 	Flocal       fs.Fs
 	Fremote      fs.Fs
 	FremoteName  string
+	Precision    time.Duration
 	cleanRemote  func()
 	mkdir        map[string]bool // whether the remote has been made yet for the fs name
 	Logf, Fatalf func(text string, args ...interface{})
@@ -89,7 +94,7 @@ func newRun() *Run {
 	Initialise()
 
 	var err error
-	r.Fremote, r.FremoteName, r.cleanRemote, err = RandomRemote(*RemoteName, *SubDir)
+	r.Fremote, r.FremoteName, r.cleanRemote, err = RandomRemote()
 	if err != nil {
 		r.Fatalf("Failed to open remote %q: %v", *RemoteName, err)
 	}
@@ -99,10 +104,13 @@ func newRun() *Run {
 		r.Fatalf("Failed to create temp dir: %v", err)
 	}
 	r.LocalName = filepath.ToSlash(r.LocalName)
-	r.Flocal, err = fs.NewFs(r.LocalName)
+	r.Flocal, err = fs.NewFs(context.Background(), r.LocalName)
 	if err != nil {
 		r.Fatalf("Failed to make %q: %v", r.LocalName, err)
 	}
+
+	r.Precision = fs.GetModifyWindow(context.Background(), r.Fremote, r.Flocal)
+
 	return r
 }
 
@@ -130,6 +138,7 @@ func retry(t *testing.T, what string, f func() error) {
 //
 // Finalise() will tidy them away when done.
 func newRunIndividual(t *testing.T, individual bool) *Run {
+	ctx := context.Background()
 	var r *Run
 	if individual {
 		r = newRun()
@@ -139,17 +148,11 @@ func newRunIndividual(t *testing.T, individual bool) *Run {
 		*r = *oneRun
 		r.cleanRemote = func() {
 			var toDelete []string
-			err := walk.Walk(r.Fremote, "", true, -1, func(dirPath string, entries fs.DirEntries, err error) error {
-				if err != nil {
-					if err == fs.ErrorDirNotFound {
-						return nil
-					}
-					t.Fatalf("Error listing: %v", err)
-				}
+			err := walk.ListR(ctx, r.Fremote, "", true, -1, walk.ListAll, func(entries fs.DirEntries) error {
 				for _, entry := range entries {
 					switch x := entry.(type) {
 					case fs.Object:
-						retry(t, fmt.Sprintf("removing file %q", x.Remote()), x.Remove)
+						retry(t, fmt.Sprintf("removing file %q", x.Remote()), func() error { return x.Remove(ctx) })
 					case fs.Directory:
 						toDelete = append(toDelete, x.Remote())
 					}
@@ -164,16 +167,18 @@ func newRunIndividual(t *testing.T, individual bool) *Run {
 			for i := len(toDelete) - 1; i >= 0; i-- {
 				dir := toDelete[i]
 				retry(t, fmt.Sprintf("removing dir %q", dir), func() error {
-					return r.Fremote.Rmdir(dir)
+					return r.Fremote.Rmdir(ctx, dir)
 				})
 			}
 			// Check remote is empty
 			CheckListingWithPrecision(t, r.Fremote, []Item{}, []string{}, r.Fremote.Precision())
+			// Clear the remote cache
+			cache.Clear()
 		}
 	}
 	r.Logf = t.Logf
 	r.Fatalf = t.Fatalf
-	r.Logf("Remote %q, Local %q, Modify Window %q", r.Fremote, r.Flocal, fs.GetModifyWindow(r.Fremote))
+	r.Logf("Remote %q, Local %q, Modify Window %q", r.Fremote, r.Flocal, fs.GetModifyWindow(ctx, r.Fremote))
 	return r
 }
 
@@ -212,7 +217,7 @@ func (r *Run) WriteFile(filePath, content string, t time.Time) Item {
 	// FIXME make directories?
 	filePath = path.Join(r.LocalName, filePath)
 	dirPath := path.Dir(filePath)
-	err := os.MkdirAll(dirPath, 0770)
+	err := file.MkdirAll(dirPath, 0770)
 	if err != nil {
 		r.Fatalf("Failed to make directories %q: %v", dirPath, err)
 	}
@@ -228,8 +233,8 @@ func (r *Run) WriteFile(filePath, content string, t time.Time) Item {
 }
 
 // ForceMkdir creates the remote
-func (r *Run) ForceMkdir(f fs.Fs) {
-	err := f.Mkdir("")
+func (r *Run) ForceMkdir(ctx context.Context, f fs.Fs) {
+	err := f.Mkdir(ctx, "")
 	if err != nil {
 		r.Fatalf("Failed to mkdir %q: %v", f, err)
 	}
@@ -237,14 +242,14 @@ func (r *Run) ForceMkdir(f fs.Fs) {
 }
 
 // Mkdir creates the remote if it hasn't been created already
-func (r *Run) Mkdir(f fs.Fs) {
+func (r *Run) Mkdir(ctx context.Context, f fs.Fs) {
 	if !r.mkdir[f.String()] {
-		r.ForceMkdir(f)
+		r.ForceMkdir(ctx, f)
 	}
 }
 
 // WriteObjectTo writes an object to the fs, remote passed in
-func (r *Run) WriteObjectTo(f fs.Fs, remote, content string, modTime time.Time, useUnchecked bool) Item {
+func (r *Run) WriteObjectTo(ctx context.Context, f fs.Fs, remote, content string, modTime time.Time, useUnchecked bool) Item {
 	put := f.Put
 	if useUnchecked {
 		put = f.Features().PutUnchecked
@@ -252,12 +257,24 @@ func (r *Run) WriteObjectTo(f fs.Fs, remote, content string, modTime time.Time,
 			r.Fatalf("Fs doesn't support PutUnchecked")
 		}
 	}
-	r.Mkdir(f)
+	r.Mkdir(ctx, f)
+
+	// calculate all hashes f supports for content
+	hash, err := hash.NewMultiHasherTypes(f.Hashes())
+	if err != nil {
+		r.Fatalf("Failed to make new multi hasher: %v", err)
+	}
+	_, err = hash.Write([]byte(content))
+	if err != nil {
+		r.Fatalf("Failed to make write to hash: %v", err)
+	}
+	hashSums := hash.Sums()
+
 	const maxTries = 10
 	for tries := 1; ; tries++ {
 		in := bytes.NewBufferString(content)
-		objinfo := object.NewStaticObjectInfo(remote, modTime, int64(len(content)), true, nil, nil)
-		_, err := put(in, objinfo)
+		objinfo := object.NewStaticObjectInfo(remote, modTime, int64(len(content)), true, hashSums, nil)
+		_, err := put(ctx, in, objinfo)
 		if err == nil {
 			break
 		}
@@ -273,19 +290,19 @@ func (r *Run) WriteObjectTo(f fs.Fs, remote, content string, modTime time.Time,
 }
 
 // WriteObject writes an object to the remote
-func (r *Run) WriteObject(remote, content string, modTime time.Time) Item {
-	return r.WriteObjectTo(r.Fremote, remote, content, modTime, false)
+func (r *Run) WriteObject(ctx context.Context, remote, content string, modTime time.Time) Item {
+	return r.WriteObjectTo(ctx, r.Fremote, remote, content, modTime, false)
 }
 
 // WriteUncheckedObject writes an object to the remote not checking for duplicates
-func (r *Run) WriteUncheckedObject(remote, content string, modTime time.Time) Item {
-	return r.WriteObjectTo(r.Fremote, remote, content, modTime, true)
+func (r *Run) WriteUncheckedObject(ctx context.Context, remote, content string, modTime time.Time) Item {
+	return r.WriteObjectTo(ctx, r.Fremote, remote, content, modTime, true)
 }
 
 // WriteBoth calls WriteObject and WriteFile with the same arguments
-func (r *Run) WriteBoth(remote, content string, modTime time.Time) Item {
+func (r *Run) WriteBoth(ctx context.Context, remote, content string, modTime time.Time) Item {
 	r.WriteFile(remote, content, modTime)
-	return r.WriteObject(remote, content, modTime)
+	return r.WriteObject(ctx, remote, content, modTime)
 }
 
 // CheckWithDuplicates does a test but allows duplicates
@@ -299,7 +316,7 @@ func (r *Run) CheckWithDuplicates(t *testing.T, items ...Item) {
 	sort.Strings(want)
 
 	// do the listing
-	objs, _, err := walk.GetAll(r.Fremote, "", true, -1)
+	objs, _, err := walk.GetAll(context.Background(), r.Fremote, "", true, -1)
 	if err != nil && err != fs.ErrorDirNotFound {
 		t.Fatalf("Error listing: %v", err)
 	}
@@ -313,6 +330,38 @@ func (r *Run) CheckWithDuplicates(t *testing.T, items ...Item) {
 	assert.Equal(t, want, got)
 }
 
+// CheckLocalItems checks the local fs with proper precision
+// to see if it has the expected items.
+func (r *Run) CheckLocalItems(t *testing.T, items ...Item) {
+	CheckItemsWithPrecision(t, r.Flocal, r.Precision, items...)
+}
+
+// CheckRemoteItems checks the remote fs with proper precision
+// to see if it has the expected items.
+func (r *Run) CheckRemoteItems(t *testing.T, items ...Item) {
+	CheckItemsWithPrecision(t, r.Fremote, r.Precision, items...)
+}
+
+// CheckLocalListing checks the local fs with proper precision
+// to see if it has the expected contents.
+//
+// If expectedDirs is non nil then we check those too.  Note that no
+// directories returned is also OK as some remotes don't return
+// directories.
+func (r *Run) CheckLocalListing(t *testing.T, items []Item, expectedDirs []string) {
+	CheckListingWithPrecision(t, r.Flocal, items, expectedDirs, r.Precision)
+}
+
+// CheckRemoteListing checks the remote fs with proper precision
+// to see if it has the expected contents.
+//
+// If expectedDirs is non nil then we check those too.  Note that no
+// directories returned is also OK as some remotes don't return
+// directories.
+func (r *Run) CheckRemoteListing(t *testing.T, items []Item, expectedDirs []string) {
+	CheckListingWithPrecision(t, r.Fremote, items, expectedDirs, r.Precision)
+}
+
 // Clean the temporary directory
 func (r *Run) cleanTempDir() {
 	err := os.RemoveAll(r.LocalName)
@@ -327,4 +376,6 @@ func (r *Run) Finalise() {
 	r.cleanRemote()
 	// r.Logf("Cleaning local %q", r.LocalName)
 	r.cleanTempDir()
+	// Clear the remote cache
+	cache.Clear()
 }
