@@ -1,5 +1,3 @@
-// +build go1.11
-
 package main
 
 import (
@@ -17,6 +15,7 @@ import (
 	"time"
 
 	"github.com/rclone/rclone/fs"
+	"github.com/rclone/rclone/lib/file"
 	"github.com/skratchdot/open-golang/open"
 )
 
@@ -77,7 +76,7 @@ func NewReport() *Report {
 
 	// Create output directory for logs and report
 	r.LogDir = path.Join(*outputDir, r.DateTime)
-	err = os.MkdirAll(r.LogDir, 0777)
+	err = file.MkdirAll(r.LogDir, 0777)
 	if err != nil {
 		log.Fatalf("Failed to make log directory: %v", err)
 	}
@@ -124,7 +123,7 @@ func (r *Report) RecordResult(t *Run) {
 	}
 }
 
-// Title returns a human readable summary title for the Report
+// Title returns a human-readable summary title for the Report
 func (r *Report) Title() string {
 	if r.AllPassed() {
 		return fmt.Sprintf("PASS: All tests finished OK in %v", r.Duration)
@@ -142,7 +141,7 @@ func (r *Report) LogSummary() {
 	if !r.AllPassed() {
 		for _, t := range r.Failed {
 			log.Printf("  * %s", toShell(t.nextCmdLine()))
-			log.Printf("    * Failed tests: %v", t.failedTests)
+			log.Printf("    * Failed tests: %v", t.FailedTests)
 		}
 	}
 }
@@ -266,7 +265,7 @@ a:focus {
 <td>{{ if ne $prevRemote .Remote }}{{ .Remote }}{{ end }}{{ $prevRemote = .Remote }}</td>
 <td>{{ .Path }}</td>
 <td><span class="{{ .FastList }}">{{ .FastList }}</span></td>
-<td>{{ .FailedTests }}</td>
+<td>{{ .FailedTestsCSV }}</td>
 <td>{{ range $i, $v := .Logs }}<a href="{{ $v }}">#{{ $i }}</a> {{ end }}</td>
 </tr>
 {{ end }}
