@@ -2,11 +2,13 @@ package vfs
 
 import (
 	"context"
+	"errors"
+	"fmt"
 	"io"
 	"os"
 	"sync"
+	"time"
 
-	"github.com/pkg/errors"
 	"github.com/rclone/rclone/fs"
 	"github.com/rclone/rclone/fs/accounting"
 	"github.com/rclone/rclone/fs/chunkedreader"
@@ -16,20 +18,21 @@ import (
 // ReadFileHandle is an open for read file handle on a File
 type ReadFileHandle struct {
 	baseHandle
-	done        func(err error)
+	done        func(ctx context.Context, err error)
 	mu          sync.Mutex
-	closed      bool // set if handle has been closed
+	cond        sync.Cond // cond lock for out of sequence reads
 	r           *accounting.Account
-	readCalled  bool  // set if read has been called
 	size        int64 // size of the object (0 for unknown length)
 	offset      int64 // offset of read of o
 	roffset     int64 // offset of Read() calls
-	noSeek      bool
-	sizeUnknown bool // set if size of source is not known
 	file        *File
 	hash        *hash.MultiHasher
-	opened      bool
 	remote      string
+	closed      bool // set if handle has been closed
+	readCalled  bool // set if read has been called
+	noSeek      bool
+	sizeUnknown bool // set if size of source is not known
+	opened      bool
 }
 
 // Check interfaces
@@ -44,8 +47,9 @@ func newReadFileHandle(f *File) (*ReadFileHandle, error) {
 	var mhash *hash.MultiHasher
 	var err error
 	o := f.getObject()
-	if !f.d.vfs.Opt.NoChecksum {
-		mhash, err = hash.NewMultiHasherTypes(o.Fs().Hashes())
+	if !f.VFS().Opt.NoChecksum {
+		hashes := hash.NewHashSet(o.Fs().Hashes().GetOne()) // just pick one hash
+		mhash, err = hash.NewMultiHasherTypes(hashes)
 		if err != nil {
 			fs.Errorf(o.Fs(), "newReadFileHandle hash error: %v", err)
 		}
@@ -53,12 +57,13 @@ func newReadFileHandle(f *File) (*ReadFileHandle, error) {
 
 	fh := &ReadFileHandle{
 		remote:      o.Remote(),
-		noSeek:      f.d.vfs.Opt.NoSeek,
+		noSeek:      f.VFS().Opt.NoSeek,
 		file:        f,
 		hash:        mhash,
 		size:        nonNegative(o.Size()),
 		sizeUnknown: o.Size() < 0,
 	}
+	fh.cond = sync.Cond{L: &fh.mu}
 	return fh, nil
 }
 
@@ -69,13 +74,13 @@ func (fh *ReadFileHandle) openPending() (err error) {
 		return nil
 	}
 	o := fh.file.getObject()
-	r, err := chunkedreader.New(context.TODO(), o, int64(fh.file.d.vfs.Opt.ChunkSize), int64(fh.file.d.vfs.Opt.ChunkSizeLimit)).Open()
+	r, err := chunkedreader.New(context.TODO(), o, int64(fh.file.VFS().Opt.ChunkSize), int64(fh.file.VFS().Opt.ChunkSizeLimit)).Open()
 	if err != nil {
 		return err
 	}
 	tr := accounting.GlobalStats().NewTransfer(o)
 	fh.done = tr.Done
-	fh.r = tr.Account(r).WithBuffer() // account the transfer
+	fh.r = tr.Account(context.TODO(), r).WithBuffer() // account the transfer
 	fh.opened = true
 
 	return nil
@@ -94,7 +99,7 @@ func (fh *ReadFileHandle) String() string {
 	return fh.file.String() + " (r)"
 }
 
-// Node returns the Node assocuated with this - satisfies Noder interface
+// Node returns the Node associated with this - satisfies Noder interface
 func (fh *ReadFileHandle) Node() Node {
 	fh.mu.Lock()
 	defer fh.mu.Unlock()
@@ -113,7 +118,7 @@ func (fh *ReadFileHandle) seek(offset int64, reopen bool) (err error) {
 	fh.hash = nil
 	if !reopen {
 		ar := fh.r.GetAsyncReader()
-		// try to fullfill the seek with buffer discard
+		// try to fulfill the seek with buffer discard
 		if ar != nil && ar.SkipBytes(int(offset-fh.offset)) {
 			fh.offset = offset
 			return nil
@@ -142,7 +147,7 @@ func (fh *ReadFileHandle) seek(offset int64, reopen bool) (err error) {
 		}
 		// re-open with a seek
 		o := fh.file.getObject()
-		r = chunkedreader.New(context.TODO(), o, int64(fh.file.d.vfs.Opt.ChunkSize), int64(fh.file.d.vfs.Opt.ChunkSizeLimit))
+		r = chunkedreader.New(context.TODO(), o, int64(fh.file.VFS().Opt.ChunkSize), int64(fh.file.VFS().Opt.ChunkSizeLimit))
 		_, err := r.Seek(offset, 0)
 		if err != nil {
 			fs.Debugf(fh.remote, "ReadFileHandle.Read seek failed: %v", err)
@@ -154,7 +159,7 @@ func (fh *ReadFileHandle) seek(offset int64, reopen bool) (err error) {
 			return err
 		}
 	}
-	fh.r.UpdateReader(r)
+	fh.r.UpdateReader(context.TODO(), r)
 	fh.offset = offset
 	return nil
 }
@@ -208,10 +213,47 @@ func (fh *ReadFileHandle) ReadAt(p []byte, off int64) (n int, err error) {
 	return fh.readAt(p, off)
 }
 
+// This waits for *poff to equal off or aborts after the timeout.
+//
+// Waits here potentially affect all seeks so need to keep them short.
+//
+// Call with fh.mu Locked
+func waitSequential(what string, remote string, cond *sync.Cond, maxWait time.Duration, poff *int64, off int64) {
+	var (
+		timeout = time.NewTimer(maxWait)
+		done    = make(chan struct{})
+		abort   = false
+	)
+	go func() {
+		select {
+		case <-timeout.C:
+			// take the lock to make sure that cond.Wait() is called before
+			// cond.Broadcast. NB cond.L == mu
+			cond.L.Lock()
+			// set abort flag and give all the waiting goroutines a kick on timeout
+			abort = true
+			fs.Debugf(remote, "aborting in-sequence %s wait, off=%d", what, off)
+			cond.Broadcast()
+			cond.L.Unlock()
+		case <-done:
+		}
+	}()
+	for *poff != off && !abort {
+		fs.Debugf(remote, "waiting for in-sequence %s to %d for %v", what, off, maxWait)
+		cond.Wait()
+	}
+	// tidy up end timer
+	close(done)
+	timeout.Stop()
+	if *poff != off {
+		fs.Debugf(remote, "failed to wait for in-sequence %s to %d", what, off)
+	}
+}
+
 // Implementation of ReadAt - call with lock held
 func (fh *ReadFileHandle) readAt(p []byte, off int64) (n int, err error) {
 	// defer log.Trace(fh.remote, "p[%d], off=%d", len(p), off)("n=%d, err=%v", &n, &err)
-	err = fh.openPending() // FIXME pending open could be more efficient in the presense of seek (and retries)
+	err = fh.openPending() // FIXME pending open could be more efficient in the presence of seek (and retries)
 	if err != nil {
 		return 0, err
 	}
@@ -220,6 +262,13 @@ func (fh *ReadFileHandle) readAt(p []byte, off int64) (n int, err error) {
 		fs.Errorf(fh.remote, "ReadFileHandle.Read error: %v", EBADF)
 		return 0, ECLOSED
 	}
+	maxBuf := 1024 * 1024
+	if len(p) < maxBuf {
+		maxBuf = len(p)
+	}
+	if gap := off - fh.offset; gap > 0 && gap < int64(8*maxBuf) {
+		waitSequential("read", fh.remote, &fh.cond, fh.file.VFS().Opt.ReadWait, &fh.offset, off)
+	}
 	doSeek := off != fh.offset
 	if doSeek && fh.noSeek {
 		return 0, ESPIPE
@@ -228,6 +277,7 @@ func (fh *ReadFileHandle) readAt(p []byte, off int64) (n int, err error) {
 	retries := 0
 	reqSize := len(p)
 	doReopen := false
+	lowLevelRetries := fs.GetConfig(context.TODO()).LowLevelRetries
 	for {
 		if doSeek {
 			// Are we attempting to seek beyond the end of the
@@ -264,11 +314,11 @@ func (fh *ReadFileHandle) readAt(p []byte, off int64) (n int, err error) {
 				break
 			}
 		}
-		if retries >= fs.Config.LowLevelRetries {
+		if retries >= lowLevelRetries {
 			break
 		}
 		retries++
-		fs.Errorf(fh.remote, "ReadFileHandle.Read error: low level retry %d/%d: %v", retries, fs.Config.LowLevelRetries, err)
+		fs.Errorf(fh.remote, "ReadFileHandle.Read error: low level retry %d/%d: %v", retries, lowLevelRetries, err)
 		doSeek = true
 		doReopen = true
 	}
@@ -291,6 +341,7 @@ func (fh *ReadFileHandle) readAt(p []byte, off int64) (n int, err error) {
 			err = io.EOF
 		}
 	}
+	fh.cond.Broadcast() // wake everyone up waiting for an in-sequence read
 	return n, err
 }
 
@@ -303,10 +354,15 @@ func (fh *ReadFileHandle) checkHash() error {
 	for hashType, dstSum := range fh.hash.Sums() {
 		srcSum, err := o.Hash(context.TODO(), hashType)
 		if err != nil {
+			if errors.Is(err, os.ErrNotExist) {
+				// if it was file not found then at
+				// this point we don't care any more
+				continue
+			}
 			return err
 		}
 		if !hash.Equals(dstSum, srcSum) {
-			return errors.Errorf("corrupted on transfer: %v hash differ %q vs %q", hashType, dstSum, srcSum)
+			return fmt.Errorf("corrupted on transfer: %v hash differ %q vs %q", hashType, dstSum, srcSum)
 		}
 	}
 
@@ -360,7 +416,7 @@ func (fh *ReadFileHandle) close() error {
 	if fh.opened {
 		var err error
 		defer func() {
-			fh.done(err)
+			fh.done(context.TODO(), err)
 		}()
 		// Close first so that we have hashes
 		err = fh.r.Close()
@@ -421,7 +477,7 @@ func (fh *ReadFileHandle) Release() error {
 	err := fh.close()
 	if err != nil {
 		fs.Errorf(fh.remote, "ReadFileHandle.Release error: %v", err)
-	} else {
+		//} else {
 		// fs.Debugf(fh.remote, "ReadFileHandle.Release OK")
 	}
 	return err
